#!/usr/bin/python3
import yaml, sqlite3
import os, re

import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib

# 创建一个DBus服务类
class KYSettingsDBusService(dbus.service.Object):
    __read_only_view = {}
    __root_view = {}
    __user_view = {}
    __user_name = None

    def __init__(self, bus_name, object_path):
        dbus.service.Object.__init__(self, bus_name, object_path)
        
        os.makedirs('/etc/kylin-config', exist_ok=True)

        # 监听SessionNew信号 创建对应用户的数据库
        self.__login_bus = dbus.SystemBus()
        self.__login_obj = self.__login_bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
        self.__login_interface = dbus.Interface(self.__login_obj,dbus_interface="org.freedesktop.login1.Manager")
        self.__login_interface.connect_to_signal('SessionNew',self.__on_SessionNew__)

        # 获取当前会话的用户名
        sessions_list = self.__login_interface.ListSessions()
        for session in sessions_list:
            obj = self.__login_bus.get_object('org.freedesktop.login1', session[4])
            interface = dbus.Interface(obj, 'org.freedesktop.DBus.Properties')
            if interface.Get('org.freedesktop.login1.Session', 'Active'):
                self.__user_name = session[2]
        self.__creat_db_file__()

        # 定时检测数据库文件 防止误删
        # self.timer = threading.Timer(600,self.__creat_db_file__)
        # self.timer.start()

    def __on_SessionNew__(self,session_id:str, object_path:dbus.ObjectPath):
        session_list = self.__login_interface.ListSessions()
        for session in session_list:
            if session[0] == session_id:
                self.__user_name = session[2]
                self.__creat_db_file__()

    def __creat_db_file__(self):
        # 系统级配置
        if not os.path.exists('/etc/kylin-config/user.db'):
            if self.__root_view == {}:
                self.__load_read_only_view__()
                self.__dict2db__(self.__read_only_view, '/etc/kylin-config/user.db')
            else:
                self.__dict2db__(self.__root_view, '/etc/kylin-config/user.db')
        if self.__root_view == {}:
            self.__root_view = self.__format_db_file('/etc/kylin-config/user.db')
        # 用户级配置 可能需要链接用户登录信号
        # if self.__user_name is not None:
        #     if not os.path.exists(f'/home/{self.__user_name}/.config/kylin-config/user.db'):
        #         if self.__user_view == {}:
        #             self.__load_read_only_view__()
        #             self.__dict2db__(self.__read_only_view, f'/home/{self.__user_name}/.config/kylin-config/user.db')
        #         else:
        #             self.__dict2db__(self.__user_view, f'/home/{self.__user_name}/.config/kylin-config/user.db')
        #     if self.__user_view == {}:
        #         self.__user_view = self.__format_db_file(f'/home/{self.__user_name}/.config/kylin-config/user.db')

    def __covert_app_table__(self,data: dict, cursor: sqlite3.Cursor):
        # 创建app表
        cursor.execute('''CREATE TABLE IF NOT EXISTS app
                        (id INTEGER PRIMARY KEY AUTOINCREMENT,
                        app_name TEXT,
                        defualt_version TEXT)''')
        #将app信息写入app表
        for app in data.keys():
            default_version = data[app].get('default_version')
            # 如果没有设置默认版本，将默认版本设置为最大版本
            if default_version is None:
                for version in data[app]:
                    if default_version is None:
                        default_version = version
                    #比较版本 版本目前认为是2.0.0-0k0格式
                    default_list = re.findall(r'\d+', default_version)
                    list = re.findall(r'\d+', version)
                    for i in range(0,6):
                        if len(default_list) == i:
                            default_version = version
                            break
                        if len(list) == i:
                            break
                        if int(default_list[i]) < int(list[i]):
                            default_version = version
                            break
                        if int(default_list[i]) > int(list[i]):
                            break
            cursor.execute("INSERT INTO app (app_name, defualt_version) VALUES (?,?)",(app, default_version))

    def __covert_version_table__(self,data: dict, cursor: sqlite3.Cursor):
        # 创建version表
        cursor.execute('''CREATE TABLE IF NOT EXISTS version
                       (id INTEGER PRIMARY KEY AUTOINCREMENT,
                       app_id INTEGER,
                       version TEXT,
                       compatible TEXT,
                       FOREIGN KEY (app_id) REFERENCES app(id))''')
        for app in data.keys():
            cursor.execute("SELECT id FROM app WHERE app_name = ?", (app,))
            appId = cursor.fetchone()
            for version in data[app].keys():
                if(version == 'default_version'):
                    continue
                cursor.execute("INSERT INTO version (app_id,version, compatible) VALUES (?, ?, ?)",(appId[0],version, data[app][version].get('compatible')))

    def __covert_configure_table__(self, data: dict, cursor: sqlite3.Cursor):
        # 创建version表
        cursor.execute('''CREATE TABLE IF NOT EXISTS configures
                        (id INTEGER PRIMARY KEY AUTOINCREMENT,
                        version_id INTEGER,
                        property TEXT,
                        group_name TEXT,
                        attr TEXT,
                        attr_type TEXT,
                        default_value TEXT,
                        permission TEXT,
                        range TEXT,
                        parent INTEGER,
                        description TEXT,
                        summary TEXT,
                        FOREIGN KEY (version_id) REFERENCES version(id),
                        FOREIGN KEY (parent) REFERENCES configures(id))''')
        cursor.execute("SELECT * FROM app")
        apps = cursor.fetchall()
        for app in apps:
            cursor.execute("SELECT * FROM version WHERE app_id = ?", (app[0],))
            versions = cursor.fetchall()
            for version in versions:
                version_data = data[app[1]][version[2]]
                #keys写入数据库
                self.__write_element_in_table__(version_data,version_data,cursor,version[0],0)

    def __recursive_search__(self, dictionary, key):
        if key in dictionary:
            return dictionary[key]
        else:
            for value in dictionary.values():
                if isinstance(value, dict):
                    result = self.__recursive_search__(value, key)
                    if result is not None:
                        return result
        return None

    def __write_element_in_table__(self, version_data:dict ,data: dict, cursor: sqlite3.Cursor, version_id, parent_id):
        #key写入到数据库
        if 'keys' in data:
            keys = data['keys']
            for key in keys:
                property = key
                attr = keys[key].get('default')
                attr_type = keys[key].get('_type')
                default_value = keys[key].get('default')
                description = keys[key].get('description')
                summary = keys[key].get('summary')
                if isinstance(default_value, str) is False:
                    default_value = str(default_value)
                #TODO: 计算range取值范围
                range = keys[key].get('range')
                # reload时，没有更新的枚举数据的range应该是枚举参数的字符串列表
                # 更新后或初始化时的枚举数据的range是@[枚举名]
                if attr_type == 'enum' and not isinstance(range,dict):
                    enum_range = {}
                    element_list = self.__recursive_search__(version_data,keys[key]['range'][1:])
                    for element in element_list:
                        enum_range[element['nick']] = int(element['value'])
                    range = str(enum_range) 
                cursor.execute('''INSERT INTO configures (version_id, property, attr, attr_type, default_value, range, parent, description, summary) 
                            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)''',(version_id, property, attr,attr_type, default_value, range, parent_id, description, summary))
        if 'children' in data:
            children = data['children']
            for child in children:
                if child.startswith('enum_'):
                    continue
                cursor.execute("INSERT INTO configures (version_id,group_name,parent) VALUES (?, ?, ?)",(version_id,child,parent_id))
                group_id = cursor.lastrowid
                self.__write_element_in_table__(version_data,children[child],cursor,version_id,group_id)

    def __load_config__(self, yaml_dir):
        if not os.path.exists(yaml_dir):
            return {}
        yaml_data = {}
        for root,dirs,files in os.walk(yaml_dir):
            for file in files:
                if file.endswith('.yaml'):
                    with open(f'{yaml_dir}/{file}','r') as tmp_file:
                        yaml_str = tmp_file.read()
                    tmp_data = yaml.safe_load(yaml_str)
                    yaml_data.update(tmp_data)
        return yaml_data
    
    def __merge_dict(self, dest:dict , src:dict):
        if src is None:
            return dest
        for key, value in src.items():
            if key in dest and isinstance(dest[key], dict) and isinstance(src[key], dict):
                self.__merge_dict(dest[key], src[key])
            else:
                dest[key] = value
        return dest

    # 合并配置视图，写数据库
    def __load_read_only_view__(self):
        self.__read_only_view.clear()
        # 只读文件目录
        self.__read_only_view.update(self.__load_config__('/etc/kylin-config/basic'))

        data = self.__load_config__('/etc/kylin-config/production')
        self.__merge_dict(self.__read_only_view, data)

        data = self.__load_config__('/etc/kylin-config/vendor')
        self.__merge_dict(self.__read_only_view, data)

        data = self.__load_config__('/etc/kylin-config/industry')
        self.__merge_dict(self.__read_only_view, data)

        data = self.__load_config__('/etc/kylin-config/project')
        self.__merge_dict(self.__read_only_view, data)

        data = self.__load_config__('/etc/kylin-config/control')
        self.__merge_dict(self.__read_only_view, data)

    def __dict2db__(self, dict, path):
        try:
            # 连接到数据库（如果不存在则创建）
            db_file_path =  path.rsplit('/',maxsplit=1)[0]
            if not os.path.exists(db_file_path):
                os.makedirs(db_file_path)
            conn = sqlite3.connect(path)

            # 创建游标对象
            cursor = conn.cursor()
            conn.execute('BEGIN TRANSACTION')

            # 获取数据库中的所有表名
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
            tables = cursor.fetchall()

            # 遍历所有表名，对每个表执行DELETE语句以删除所有行
            for table in tables:
                cursor.execute(f"DELETE FROM {table[0]}")

            self.__covert_app_table__(dict,cursor)
            self.__covert_version_table__(dict,cursor)
            self.__covert_configure_table__(dict,cursor)

            # 提交更改
            conn.commit()
            conn.close()
            return 0
        
        except sqlite3.Error as e:
            conn.rollback()
            conn.close()
            return -1

    #读取一个元素
    def __format_element(self, parent: dict, cursor: sqlite3.Cursor, version_id, parent_id):
        keys = {}
        children = {}
        cursor.execute('SELECT * FROM configures WHERE version_id = ? AND parent = ?', (version_id,parent_id))
        configure_data = cursor.fetchall()
        for configure in configure_data:
            #读取子分组
            if configure[3] is not None:
                children[configure[3]] = {}
                self.__format_element(children[configure[3]],cursor,version_id,configure[0])
            #读取key
            else:
                keys[configure[2]] = {}
                if configure[2] is not None:
                    keys[configure[2]]['_property'] = configure[2]
                if configure[4] is not None:
                    keys[configure[2]]['_type'] = configure[4]
                if configure[5] is not None:
                    keys[configure[2]]['default'] = configure[5]
                if configure[6] is not None:
                    keys[configure[2]]['default'] = configure[6]
                if configure[7] is not None:
                    keys[configure[2]]['permission'] = configure[7]
                if configure[8] is not None:
                    if configure[4] == 'enum':
                        type = GLib.VariantType.new('a{si}')
                        variant = GLib.Variant.parse(type,configure[9])
                        range = {}
                        range.update(variant)
                        keys[configure[2]]['range'] = range
                    else:
                        keys[configure[2]]['range'] = configure[8]
                if configure[10] is not None:
                    keys[configure[2]]['description'] = configure[10]
                if configure[11] is not None:
                    keys[configure[2]]['summary'] = configure[11]
        if keys:
            parent['keys'] = keys
        if children:
            parent['children'] = children            

    #将数据库中的内容读取到数据结构中 
    #由于存储数组到数据库不会存储 枚举 所以反序列化出来的数据结构中不存在 枚举
    def __format_db_file(self, db_file: str):
        data = {}
        # 连接到数据库（如果不存在则创建）
        conn = sqlite3.connect(db_file)
        # 创建游标对象
        cursor = conn.cursor()

        #遍历app表
        cursor.execute('SELECT * FROM app')
        app_data = cursor.fetchall()
        for app in app_data:
            data[app[1]] = {}
            cursor.execute('SELECT * FROM version WHERE app_id = ?', (app[0],))
            version_data = cursor.fetchall()
            for version in version_data:
                data[app[1]][version[2]] = {}
                self.__format_element(data[app[1]][version[2]],cursor,version[0],0)
        conn.close()
        return data

    # key值改变信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='sss')
    def key_changed(self, id, version, key):
        pass

    # 配置信息update信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='')
    def updated(self):
        pass

    # 设置一个配置的值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='sssss', out_signature='i')
    def set(self, user, id, version, key, value):
        try:
            # 修改用户配置文件，写入到数据库
            group_list = id.split('.')
            sql_path = ''
            view = {}
            if user == 'root':
                view = self.__root_view
                sql_path = '/etc/kylin-config/user.db'
            else:
                view = self.__user_view
                sql_path = f'/home/{user}/.config/kylin-config/user.db'

            conn = sqlite3.connect(sql_path)
            # 创建游标对象
            cursor = conn.cursor()

            # 开始写入事务
            conn.execute('BEGIN TRANSACTION')
            cursor.execute("SELECT id FROM app WHERE app_name = ?", (group_list[0],))
            app_id = cursor.fetchone()[0]
            cursor.execute("SELECT id FROM version WHERE app_id = ? AND version = ?", (app_id,version))
            version_id = cursor.fetchone()[0]
            parent_id = 0
            # 查找组
            for group_name in group_list[1:]:
                cursor.execute("SELECT id FROM configures WHERE version_id = ? AND parent = ? AND group_name = ?", (version_id, parent_id, group_name))
                parent_id = cursor.fetchone()[0]
            # cursor.execute("SELECT * FROM configures WHERE version_id = ? AND parent = ? AND property = ?", (version_id, parent_id, key))
            # key_data = cursor.fetchone()

            cursor.execute("UPDATE configures SET attr = ? WHERE version_id = ? AND parent = ? AND property = ?", (value,version_id, parent_id, key))
            conn.commit()
            conn.close()
            view.update(self.__format_db_file(sql_path))
            self.key_changed(id, version, key)
            return 1
        except sqlite3.Error as e:
            conn.rollback()
            conn.close()
            return 0
    
    # 恢复单个默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ssss', out_signature='i')
    def reset(self,user, id, version, key):
        try:
            # 修改用户配置文件，写入到数据库
            group_list = id.split('.')
            sql_path = ''
            view = {}
            if user == 'root':
                view = self.__root_view
                sql_path = '/etc/kylin-config/user.db'
            else:
                view = self.__user_view
                sql_path = f'/home/{user}/.config/kylin-config/user.db'

            conn = sqlite3.connect(sql_path)
            # 创建游标对象
            cursor = conn.cursor()

            # 开始写入事务
            conn.execute('BEGIN TRANSACTION')
            cursor.execute("SELECT id FROM app WHERE app_name = ?", (group_list[0],))
            app_id = cursor.fetchone()[0]
            cursor.execute("SELECT id FROM version WHERE app_id = ? AND version = ?", (app_id,version))
            version_id = cursor.fetchone()[0]
            parent_id = 0
            # 查找组
            for group_name in group_list[1:]:
                cursor.execute("SELECT id FROM configures WHERE version_id = ? AND parent = ? AND group_name = ?", (version_id, parent_id, group_name))
                parent_id = cursor.fetchone()[0]
            # cursor.execute("SELECT * FROM configures WHERE version_id = ? AND parent = ? AND property = ?", (version_id, parent_id, key))
            cursor.execute("UPDATE configures SET attr = default_value WHERE version_id = ? AND parent = ? AND property = ?", (version_id, parent_id, key))
            conn.commit()
            conn.close()
            view.update(self.__format_db_file(sql_path))
            self.key_changed(id,version,key)
            return 0
        except sqlite3.Error as e:
            conn.rollback()
            conn.close()
            return -1
    
    # 恢复整个id的默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='sss', out_signature='i')
    def reset_recursively(self, user, id, version):
        try:
            # 修改用户配置文件，写入到数据库
            group_list = id.split('.')
            view = {}
            if user == 'root':
                view = self.__root_view
                sql_path = '/etc/kylin-config/user.db'
            else:
                view = self.__user_view
                sql_path = f'/home/{user}/.config/kylin-config/user.db'

            conn = sqlite3.connect(sql_path)
            # 创建游标对象
            cursor = conn.cursor()

            # 开始写入事务
            conn.execute('BEGIN TRANSACTION')
            cursor.execute("SELECT id FROM app WHERE app_name = ?", (group_list[0],))
            app_id = cursor.fetchone()[0]
            cursor.execute("SELECT id FROM version WHERE app_id = ? AND version = ?", (app_id,version))
            version_id = cursor.fetchone()[0]
            parent_id = 0
            # 查找组
            for group_name in group_list[1:]:
                cursor.execute("SELECT id FROM configures WHERE version_id = ? AND parent = ? AND group_name = ?", (version_id, parent_id, group_name))
                parent_id = cursor.fetchone()[0]
            cursor.execute("UPDATE configures SET value = default_value WHERE version_id = ? AND parent = ?", (version_id, parent_id))
            conn.commit()
            conn.close()
            view.update(self.__format_db_file(sql_path))
            self.key_changed(id,version,None)
            return 0
        except sqlite3.Error as e:
            conn.rollback()
            conn.close()
            return -1

    # 恢复整个id的默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='', out_signature='i')
    def reload(self):
        self.__load_read_only_view__()
        self.__merge_dict(self.__root_view, self.__read_only_view)
        self.__merge_dict(self.__user_view, self.__read_only_view)

        if 0 == self.__dict2db__(self.__root_view, '/etc/kylin-config/user.db'):
            pass
        if 0 == self.__dict2db__(self.__user_view, f'/home/{self.__user_name}/.config/kylin-config/user.db'):
            pass
        self.updated()
        return 1

if __name__ == "__main__":
    # 初始化DBus主循环
    DBusGMainLoop(set_as_default=True)

    # 连接到DBus系统总线
    bus = dbus.SystemBus()

    # 注册一个DBus服务
    bus_name = dbus.service.BusName("com.kylin.kysdk.conf2", bus=bus)
    object_path = "/com/kylin/kysdk/conf2"
    my_dbus_service = KYSettingsDBusService(bus_name, object_path)

    # 开始DBus主循环
    mainloop = GLib.MainLoop()
    mainloop.run()